//--------------------------------------------------------------------------------------
// File: DX10 Heightfield Demo.cpp
//
// This sample shows a simple example of the Microsoft Direct3D's High-Level 
// Shader Language (HLSL) using the Effect interface. 
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//--------------------------------------------------------------------------------------
#include "DXUT.h"
#include "DXUTcamera.h"
#include "DXUTgui.h"
#include "DXUTsettingsDlg.h"
#include "SDKmisc.h"
#include "resource.h"
#include "Heightfield.h"

//--------------------------------------------------------------------------------------
// UI control IDs
//--------------------------------------------------------------------------------------

#define IDC_TOGGLEFULLSCREEN    1
#define IDC_TOGGLEREF           3
#define IDC_CHANGEDEVICE        4
#define IDC_TOGGLEWIREFRAME     6
#define IDC_TOGGLEANIMATING		7
#define IDC_RESOLUTION_STATIC	8
#define IDC_RESOLUTION_SLIDER	9
#define IDC_NORMAL_MODE			10
#define DEFAULT_RESOLUTION		127 //3.5

//--------------------------------------------------------------------------------------
// Global variables
//--------------------------------------------------------------------------------------

ID3DX10Font*                g_pFont10		= NULL;
ID3DX10Sprite*              g_pSprite10		= NULL;
ID3D10Effect*               g_pEffect		= NULL;
ID3D10InputLayout*          g_pVertexLayout = NULL;
Heightfield*				g_pHeightfield	= NULL;

CModelViewerCamera			g_Camera;					// A model viewing camera
CDXUTDirectionWidget		g_LightControl;
CDXUTDialogResourceManager  g_DialogResourceManager;	// manager for shared resources of dialogs
CD3DSettingsDlg             g_D3DSettingsDlg;			// Device settings dialog
CDXUTDialog                 g_HUD;						// manages the 3D   
CDXUTDialog                 g_SampleUI;					// dialog for sample specific controls
bool                        g_bShowHelp		= false;    // If true, it renders the UI control text
bool						g_IsWireframe   = false;
bool						g_IsAnimating	= true;
NormalMode					g_NormalMode	= NormalMode_FilterNormals;

ID3D10EffectVectorVariable*	g_pLightDir				= NULL;
ID3D10EffectVectorVariable*	g_pLightDiffuse			= NULL;
ID3D10EffectMatrixVariable*	g_pViewProjection		= NULL;

//--------------------------------------------------------------------------------------
// Forward declarations 
//--------------------------------------------------------------------------------------

LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing, void* pUserContext );
HRESULT CALLBACK OnCreateDevice( ID3D10Device* pd3dDevice, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );
HRESULT CALLBACK OnResizedSwapChain( ID3D10Device* pd3dDevice, IDXGISwapChain* pSwapChain, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );
bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, void* pUserContext );
void CALLBACK OnFrameMove( double fTime, float fElapsedTime, void* pUserContext );
void CALLBACK OnKeyboard( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext );
void CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext );
bool CALLBACK IsDeviceAcceptable( UINT Adapter, UINT Output, D3D10_DRIVER_TYPE DeviceType, DXGI_FORMAT BackBufferFormat, bool bWindowed, void* pUserContext );
void CALLBACK OnReleasingSwapChain( void* pUserContext );
void CALLBACK OnDestroyDevice( void* pUserContext );
void CALLBACK OnFrameRender( ID3D10Device* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext );

void InitApp();
void RenderText();


//--------------------------------------------------------------------------------------
// Entry point to the program. Initializes everything and goes into a message processing 
// loop. Idle time is used to render the scene.
//--------------------------------------------------------------------------------------
int WINAPI wWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nCmdShow )
{
    // Enable run-time memory check for debug builds.
#if defined(DEBUG) | defined(_DEBUG)
    _CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif

    // DXUT will create and use the best device (either D3D9 or D3D10) 
    // that is available on the system depending on which D3D callbacks are set below

    // Set DXUT callbacks
    DXUTSetCallbackDeviceChanging( ModifyDeviceSettings );
    DXUTSetCallbackMsgProc( MsgProc );
    DXUTSetCallbackKeyboard( OnKeyboard );
    DXUTSetCallbackFrameMove( OnFrameMove );
    DXUTSetCallbackD3D10DeviceAcceptable( IsDeviceAcceptable );
    DXUTSetCallbackD3D10DeviceCreated( OnCreateDevice );
    DXUTSetCallbackD3D10SwapChainResized( OnResizedSwapChain );
    DXUTSetCallbackD3D10FrameRender( OnFrameRender );
    DXUTSetCallbackD3D10SwapChainReleasing( OnReleasingSwapChain );
    DXUTSetCallbackD3D10DeviceDestroyed( OnDestroyDevice );

    InitApp();
    DXUTInit(); 
    DXUTSetCursorSettings( true, true ); // Show the cursor and clip it when in full screen
    DXUTCreateWindow( L"DX10 Heightfield Demo" );
    DXUTCreateDevice( true, 800, 600 );
    DXUTMainLoop(); 

    return DXUTGetExitCode();
}


//--------------------------------------------------------------------------------------
// Initialize the app 
//--------------------------------------------------------------------------------------
void InitApp()
{
    g_LightControl.SetLightDirection( D3DXVECTOR3(  -.577f, .577f, -.577f ) );

    // Initialize dialogs
    g_D3DSettingsDlg.Init( &g_DialogResourceManager );
    g_HUD.Init( &g_DialogResourceManager );
    g_SampleUI.Init( &g_DialogResourceManager );

    g_HUD.SetCallback( OnGUIEvent ); int iY = 10;
    g_HUD.AddButton( IDC_TOGGLEFULLSCREEN, L"Toggle full screen", 35, iY, 125, 22 );
    g_HUD.AddButton( IDC_TOGGLEREF, L"Toggle REF (F3)", 35, iY += 24, 125, 22, VK_F3 );
    g_HUD.AddButton( IDC_CHANGEDEVICE, L"Change device (F2)", 35, iY += 24, 125, 22, VK_F2 );

    g_SampleUI.SetCallback( OnGUIEvent );

	iY = 100;
	g_SampleUI.AddStatic( IDC_RESOLUTION_STATIC, L"", 35, iY += 24, 125, 22, false );
	g_SampleUI.AddSlider( IDC_RESOLUTION_SLIDER, 35, iY += 24, 125, 22, 16, 511, DEFAULT_RESOLUTION, true );
	
	CDXUTComboBox* pComboBox;
	g_SampleUI.AddComboBox(IDC_NORMAL_MODE, 10, iY += 30, 150, 22, 0, false, &pComboBox );
	pComboBox->AddItem(L"Filtered Normals", (void*)NormalMode_FilterNormals );
	pComboBox->AddItem(L"CPU Normals",		(void*)NormalMode_CpuNormals );
	pComboBox->AddItem(L"GPU Normals",		(void*)NormalMode_GpuNormals );
	pComboBox->AddItem(L"Streamed Normals", (void*)NormalMode_StreamNormals );

	g_SampleUI.AddButton( IDC_TOGGLEANIMATING, L"Toggle Animating", 35, iY += 30, 125, 22 );
	g_SampleUI.AddButton( IDC_TOGGLEWIREFRAME, L"Toggle Wireframe", 35, iY += 30, 125, 22 );

	WCHAR sz[100];
	StringCchPrintf( sz, 100, L"Resolution: %d", g_SampleUI.GetSlider( IDC_RESOLUTION_SLIDER )->GetValue() + 1 );
	g_SampleUI.GetStatic( IDC_RESOLUTION_STATIC )->SetText( sz );
}


//--------------------------------------------------------------------------------------
// Called right before creating a D3D9 or D3D10 device, allowing the app to modify the device settings as needed
//--------------------------------------------------------------------------------------

bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, void* pUserContext )
{
    // Uncomment this to get debug information from D3D10
    //pDeviceSettings->d3d10.CreateFlags |= D3D10_CREATE_DEVICE_DEBUG;

    // For the first device created if its a REF device, optionally display a warning dialog box
    static bool s_bFirstTime = true;
    if( s_bFirstTime )
    {
        s_bFirstTime = false;
        if( DXUT_D3D10_DEVICE == pDeviceSettings->ver &&
            pDeviceSettings->d3d10.DriverType == D3D10_DRIVER_TYPE_REFERENCE )
            DXUTDisplaySwitchingToREFWarning( pDeviceSettings->ver );
    }

    return true;
}


//--------------------------------------------------------------------------------------
// Handle updates to the scene.  This is called regardless of which D3D API is used
//--------------------------------------------------------------------------------------

void CALLBACK OnFrameMove( double fTime, float fElapsedTime, void* pUserContext )
{
    //Update the camera's position based on user input 
	g_Camera.FrameMove( fElapsedTime );

	if( g_IsAnimating )
		g_pHeightfield->Update(fElapsedTime);
}


//--------------------------------------------------------------------------------------
// Render the help and statistics text
//--------------------------------------------------------------------------------------

void RenderText()
{
	CDXUTTextHelper txtHelper(g_pFont10, g_pSprite10 );

    txtHelper.Begin();
    txtHelper.SetInsertionPos( 2, 0 );
    txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 0.0f, 1.0f ) );
    txtHelper.DrawTextLine( DXUTGetFrameStats( false ) );
    txtHelper.DrawTextLine( DXUTGetDeviceStats() );
	txtHelper.DrawFormattedTextLine( L"FPS: %0.2f  SPF: %0.8f", DXUTGetFPS(), 1 / DXUTGetFPS() );

    // Draw help
    if( g_bShowHelp )
    {
        UINT nBackBufferHeight = DXUTGetDXGIBackBufferSurfaceDesc()->Height;
        txtHelper.SetInsertionPos( 2, nBackBufferHeight - 15 * 5 );
        txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 0.75f, 0.0f, 1.0f ) );

        txtHelper.DrawTextLine( L"Controls:" );
        txtHelper.SetInsertionPos( 20, nBackBufferHeight - 15 * 4 );
        txtHelper.DrawTextLine( L"Rotate light: Right mouse button\n"
                                    L"Rotate camera: Middle mouse button\n"
                                    L"Zoom camera: Mouse wheel scroll\n" );

        txtHelper.SetInsertionPos( 250, nBackBufferHeight - 15 * 4 );
        txtHelper.DrawTextLine( L"Hide help: F1\n"
                                    L"Quit: ESC\n" );
    }
    else
    {
        txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 1.0f, 1.0f ) );
        txtHelper.DrawTextLine( L"Press F1 for help" );
    }

    txtHelper.End();
}


//--------------------------------------------------------------------------------------
// Handle messages to the application
//--------------------------------------------------------------------------------------

LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing,
                          void* pUserContext )
{
    // Pass messages to dialog resource manager calls so GUI state is updated correctly
    *pbNoFurtherProcessing = g_DialogResourceManager.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;

    // Pass messages to settings dialog if its active
    if( g_D3DSettingsDlg.IsActive() )
    {
        g_D3DSettingsDlg.MsgProc( hWnd, uMsg, wParam, lParam );
        return 0;
    }

    // Give the dialogs a chance to handle the message first
    *pbNoFurtherProcessing = g_HUD.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;
    *pbNoFurtherProcessing = g_SampleUI.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;

    g_LightControl.HandleMessages( hWnd, uMsg, wParam, lParam );

    // Pass all remaining windows messages to camera so it can respond to user input
    g_Camera.HandleMessages( hWnd, uMsg, wParam, lParam );

    return 0;
}


//--------------------------------------------------------------------------------------
// Handle key presses
//--------------------------------------------------------------------------------------

void CALLBACK OnKeyboard( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext )
{
    if( bKeyDown && nChar == VK_F1 )
		g_bShowHelp = !g_bShowHelp;
}


//--------------------------------------------------------------------------------------
// Handles the GUI events
//--------------------------------------------------------------------------------------

void CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext )
{
    switch( nControlID )
    {
        case IDC_TOGGLEFULLSCREEN:
            DXUTToggleFullScreen(); break;
        case IDC_TOGGLEREF:
            DXUTToggleREF(); break;
        case IDC_CHANGEDEVICE:
            g_D3DSettingsDlg.SetActive( !g_D3DSettingsDlg.IsActive() ); break;

		case IDC_NORMAL_MODE:
			{
				int mode = (int)g_SampleUI.GetComboBox(IDC_NORMAL_MODE)->GetSelectedData();
				g_NormalMode = (NormalMode)mode;
				g_pHeightfield->Mode = g_NormalMode;
				g_pHeightfield->Update(0);
			}
			break;

		case IDC_TOGGLEWIREFRAME:
		{
			g_IsWireframe = !g_IsWireframe;
			g_pHeightfield->IsWireframe = g_IsWireframe;
			break;
		}

		case IDC_TOGGLEANIMATING:
		{
			g_IsAnimating = !g_IsAnimating;
			break;
		}

		case IDC_RESOLUTION_SLIDER:
		{
			int value = g_SampleUI.GetSlider( IDC_RESOLUTION_SLIDER )->GetValue();
			SAFE_DELETE(g_pHeightfield);
			g_pHeightfield = new Heightfield();
			g_pHeightfield->Init(16.0f / value, DXUTGetD3D10Device(), g_pEffect );
			g_pHeightfield->Mode		= g_NormalMode;
			g_pHeightfield->IsWireframe = g_IsWireframe;
			g_pHeightfield->Update(0);
			
			WCHAR sz[100];
			StringCchPrintf( sz, 100, L"Resolution: %d", g_SampleUI.GetSlider( IDC_RESOLUTION_SLIDER )->GetValue() + 1 );
			g_SampleUI.GetStatic( IDC_RESOLUTION_STATIC )->SetText( sz );
		}
    }
}


//--------------------------------------------------------------------------------------
// Reject any D3D10 devices that aren't acceptable by returning false
//--------------------------------------------------------------------------------------

bool CALLBACK IsDeviceAcceptable( UINT Adapter, UINT Output, D3D10_DRIVER_TYPE DeviceType,
                                       DXGI_FORMAT BackBufferFormat, bool bWindowed, void* pUserContext )
{
    return true;
}


//--------------------------------------------------------------------------------------
// Create any D3D10 resources that aren't dependant on the back buffer
//--------------------------------------------------------------------------------------

HRESULT CALLBACK OnCreateDevice( ID3D10Device* pd3dDevice, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc,
                                      void* pUserContext )
{
    HRESULT hr;

    V_RETURN( g_DialogResourceManager.OnD3D10CreateDevice( pd3dDevice ) );
    V_RETURN( g_D3DSettingsDlg.OnD3D10CreateDevice( pd3dDevice ) );
    V_RETURN( D3DX10CreateFont( pd3dDevice, 15, 0, FW_BOLD, 1, FALSE, DEFAULT_CHARSET,
                                OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE,
                                L"Arial", &g_pFont10 ) );

    V_RETURN( D3DX10CreateSprite( pd3dDevice, 512, &g_pSprite10 ) );
	
    V_RETURN( CDXUTDirectionWidget::StaticOnD3D10CreateDevice( pd3dDevice ) );
	g_LightControl.SetRadius( 1.0f );

    DWORD dwShaderFlags = D3D10_SHADER_ENABLE_STRICTNESS;

#if defined( DEBUG ) || defined( _DEBUG )
    // Set the D3D10_SHADER_DEBUG flag to embed debug information in the shaders.
    // Setting this flag improves the shader debugging experience, but still allows 
    // the shaders to be optimized and to run exactly the way they will run in 
    // the release configuration of this program.
    dwShaderFlags |= D3D10_SHADER_DEBUG;
#endif

    // Read the D3DX effect file
    WCHAR str[MAX_PATH];
    V_RETURN( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"DX10 Heightfield Demo.fx" ) );
    V_RETURN( D3DX10CreateEffectFromFile( str, NULL, NULL, "fx_4_0", dwShaderFlags, 0, pd3dDevice, NULL,
                                          NULL, &g_pEffect, NULL, NULL ) );    

	if( g_pHeightfield ) delete g_pHeightfield;
	g_pHeightfield = new Heightfield();
	g_pHeightfield->Init( 16.0f / DEFAULT_RESOLUTION, pd3dDevice, g_pEffect );
	g_pHeightfield->Mode		= g_NormalMode;
	g_pHeightfield->IsWireframe = g_IsWireframe;

    // Setup the camera's view parameters
	D3DXVECTOR4 vLightDiffuse(1,1,1,1); 
    D3DXVECTOR3 vecEye( 8.0f, 8.0f, -8.0f );
    D3DXVECTOR3 vecAt ( 8.0f, 0.0f, 5.0f );
    g_Camera.SetViewParams( &vecEye, &vecAt );
    g_Camera.SetRadius( 22.0f );

	// Obtain variables
	g_pLightDir			= g_pEffect->GetVariableByName( "g_LightDir" )->AsVector();
    g_pLightDiffuse		= g_pEffect->GetVariableByName( "g_LightDiffuse" )->AsVector();
    g_pViewProjection	= g_pEffect->GetVariableByName( "g_ViewProjectionMatrix" )->AsMatrix();
	
	V( g_pLightDiffuse->SetFloatVector( ( float* )&vLightDiffuse ));

    return S_OK;
}


//--------------------------------------------------------------------------------------
// Create any D3D10 resources that depend on the back buffer
//--------------------------------------------------------------------------------------

HRESULT CALLBACK OnResizedSwapChain( ID3D10Device* pd3dDevice, IDXGISwapChain* pSwapChain,
                            const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
    HRESULT hr;
    V_RETURN( g_DialogResourceManager.OnD3D10ResizedSwapChain( pd3dDevice, pBackBufferSurfaceDesc ) );
    V_RETURN( g_D3DSettingsDlg.OnD3D10ResizedSwapChain( pd3dDevice, pBackBufferSurfaceDesc ) );

    // Setup the camera's projection parameters
    float fAspectRatio = pBackBufferSurfaceDesc->Width / ( FLOAT )pBackBufferSurfaceDesc->Height;
    g_Camera.SetProjParams( D3DX_PI / 4, fAspectRatio, 2.0f, 4000.0f );
    g_Camera.SetWindow( pBackBufferSurfaceDesc->Width, pBackBufferSurfaceDesc->Height );
    g_Camera.SetButtonMasks( MOUSE_LEFT_BUTTON, MOUSE_WHEEL, MOUSE_MIDDLE_BUTTON );

    g_HUD.SetLocation( pBackBufferSurfaceDesc->Width - 170, 0 );
    g_HUD.SetSize( 170, 170 );
    g_SampleUI.SetLocation( pBackBufferSurfaceDesc->Width - 170, pBackBufferSurfaceDesc->Height - 300 );
    g_SampleUI.SetSize( 170, 300 );

    return S_OK;
}


//--------------------------------------------------------------------------------------
// Render the scene using the D3D10 device
//--------------------------------------------------------------------------------------

void CALLBACK OnFrameRender( ID3D10Device* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext )
{
    HRESULT hr;

    // If the settings dialog is being shown, then render it instead of rendering the app's scene
    if( g_D3DSettingsDlg.IsActive() )
    {
        g_D3DSettingsDlg.OnRender( fElapsedTime );
        return;
    }

    // Clear the render target and depth stencil
    float ClearColor[4] = { 0.22f, 0.58f, 0.93f, 1.0f };
    ID3D10RenderTargetView* pRTV = DXUTGetD3D10RenderTargetView();
	ID3D10DepthStencilView* pDSV = DXUTGetD3D10DepthStencilView();
    pd3dDevice->ClearRenderTargetView( pRTV, ClearColor );
    pd3dDevice->ClearDepthStencilView( pDSV, D3D10_CLEAR_DEPTH, 1.0, 0 );
	
	D3DXMATRIX mView			= *g_Camera.GetViewMatrix();
    D3DXMATRIX mProj			= *g_Camera.GetProjMatrix();
	D3DXCOLOR arrowColor		= D3DXVECTOR4( 1, 1, 0, 1 );
	V( g_LightControl.OnRender10( arrowColor, &mView, &mProj, g_Camera.GetEyePt() ) );

	D3DXMATRIX viewProjection	= mView * mProj;    
	V( g_pViewProjection->SetMatrix( ( float* )&viewProjection ) );

	D3DXVECTOR3 vLightDir		= g_LightControl.GetLightDirection();
    V( g_pLightDir->SetRawValue( &vLightDir, 0, sizeof( D3DXVECTOR3 )) );	
    
	g_pHeightfield->Draw();
	
    g_HUD.OnRender( fElapsedTime );
    g_SampleUI.OnRender( fElapsedTime );
    RenderText();
}


//--------------------------------------------------------------------------------------
// Release D3D10 resources created in OnD3D10ResizedSwapChain 
//--------------------------------------------------------------------------------------

void CALLBACK OnReleasingSwapChain( void* pUserContext )
{
    g_DialogResourceManager.OnD3D10ReleasingSwapChain();
}


//--------------------------------------------------------------------------------------
// Release D3D10 resources created in OnD3D10CreateDevice 
//--------------------------------------------------------------------------------------
void CALLBACK OnDestroyDevice( void* pUserContext )
{
    g_DialogResourceManager.OnD3D10DestroyDevice();
    g_D3DSettingsDlg.OnD3D10DestroyDevice();
    CDXUTDirectionWidget::StaticOnD3D10DestroyDevice();
    DXUTGetGlobalResourceCache().OnDestroyDevice();

    SAFE_RELEASE( g_pFont10 );
    SAFE_RELEASE( g_pSprite10 );
    SAFE_RELEASE( g_pEffect );
    SAFE_RELEASE( g_pVertexLayout );
	SAFE_RELEASE( g_pHeightfield );
}



