//--------------------------------------------------------------------------------------
// File: DX10 Heightfield Demo.fx
//--------------------------------------------------------------------------------------

//--------------------------------------------------------------------------------------
// Output Structures
//--------------------------------------------------------------------------------------

struct VS_OUTPUT
{
    float4 Position : SV_POSITION;
    float4 Diffuse  : COLOR0;
};

struct PS_OUTPUT
{
    float4 RGBColor : SV_Target; // Pixel color
};

struct GS_INPUT
{
	float4 Position : POSITION;
};

//--------------------------------------------------------------------------------------
// Constants & Buffers
//--------------------------------------------------------------------------------------

cbuffer cbNeverChanges
{
	float3	g_LightDiffuse;
	float3	g_MaterialDiffuseColor;
};

cbuffer cbMouseChanged
{
	float3		g_LightDir;	
	float4x4	g_ViewProjectionMatrix;
};

cbuffer cbOptionChanged
{
	float	g_MetersPerVertex;	// The spacing between vertices
	int		g_NumVertsWide;
	int		g_NumVertsDeep;
};

Buffer<float3> g_SurfaceNormals;
Buffer<float>  g_Heights;

//--------------------------------------------------------------------------------------
// States
//--------------------------------------------------------------------------------------

DepthStencilState EnableDepth
{
    DepthEnable		= TRUE;
    DepthWriteMask	= ALL;
    DepthFunc		= LESS_EQUAL;
};


//--------------------------------------------------------------------------------------
// Compute Normals
//--------------------------------------------------------------------------------------

float3 ComputeNormal( uint index )
{
	float3 normal 		= 0.0;	
	int topVertex		= g_NumVertsDeep - 1;
	int rightVertex		= g_NumVertsWide - 1;	
	int normalsPerRow   = rightVertex * 2;
	int numRows			= topVertex;	
	float top			= normalsPerRow * (numRows - 1);

	int x = index % g_NumVertsWide;
	int z = index / g_NumVertsWide;
	
	// Bottom
	if( z == 0 )
	{
		if( x == 0 )
		{
			float3 normal0 = g_SurfaceNormals.Load( int2( 0, 0 ) );
			float3 normal1 = g_SurfaceNormals.Load( int2( 1, 0 ) );
			normal = normal0 + normal1;
		}
		else if(x == rightVertex)
		{
			index = ( normalsPerRow - 1 );
			normal = g_SurfaceNormals.Load( int2( index, 0 ) );
		}
		else
		{
			index = (2 * x);
			normal = g_SurfaceNormals.Load( int2( index-1, 0 ) ) +
			         g_SurfaceNormals.Load( int2( index,   0 ) ) +
			         g_SurfaceNormals.Load( int2( index+1, 0 ) );
		}
	}
	
	// Top
	else if( z == topVertex )
	{
		if(x == 0)
		{			
			normal = g_SurfaceNormals.Load( int2( top, 0 ) );
		}
		else if(x == rightVertex)
		{
			index =	(normalsPerRow * numRows) - 1;
			normal = g_SurfaceNormals.Load( int2( index,   0 ) ) +
			         g_SurfaceNormals.Load( int2( index-1, 0 ) );
		}
		else
		{	

			index = top + (2 * x);
			normal = g_SurfaceNormals.Load( int2( index-2, 0) ) +
			         g_SurfaceNormals.Load( int2( index,   0) ) +
			         g_SurfaceNormals.Load( int2( index-1, 0) );
		}
	}
	
	// Middle
	else
	{
		if(x == 0)
		{
			int index1 = z * normalsPerRow;
			int index2 = index1 - normalsPerRow;
			normal = g_SurfaceNormals.Load( int2( index1,   0 ) ) +
			         g_SurfaceNormals.Load( int2( index1+1, 0 ) ) +
			         g_SurfaceNormals.Load( int2( index2,   0 ) );
		}
		else if(x == rightVertex)
		{
			int index1 = ( z + 1 ) * normalsPerRow - 1;
			int index2 = index1 - normalsPerRow;
			normal = g_SurfaceNormals.Load( int2( index1,   0 ) ) +
			         g_SurfaceNormals.Load( int2( index2,	  0 ) ) +
			         g_SurfaceNormals.Load( int2( index2-1, 0 ) );
		}
		else
		{
			int index1 = ( z * normalsPerRow ) + ( 2 * x );
			int index2 = index1 - normalsPerRow;
			normal = g_SurfaceNormals.Load( int2( index1-1, 0 ) ) +
			         g_SurfaceNormals.Load( int2( index1,   0 ) ) +
			         g_SurfaceNormals.Load( int2( index1+1, 0 ) ) +					 
			         g_SurfaceNormals.Load( int2( index2-2, 0 ) ) +
			         g_SurfaceNormals.Load( int2( index2-1, 0 ) ) +
			         g_SurfaceNormals.Load( int2( index2,   0 ) );
		}
	}	

	return normal;
}

float Height(int index)
{
	return g_Heights.Load( int2( index, 0 ) );	
}

float3 FilterNormal( float2 pos, int index )
{
	float3 normal = float3( 0, 1, 0);

	if( pos.y == 0)
		normal.z = Height( index ) - Height( index + g_NumVertsWide );

	else if( pos.y == g_NumVertsDeep - 1 )
		normal.z = Height( index - g_NumVertsWide ) - Height( index );
	
	else
		normal.z =  (( Height( index ) - Height( index + g_NumVertsWide ) ) +
					 ( Height( index - g_NumVertsWide ) - Height( index ) )) * 0.5;

	if( pos.x == 0)
		normal.x = Height( index ) - Height( index + 1 );
	
	else if( pos.x == g_NumVertsWide - 1 )
		normal.x = Height( index - 1 ) - Height( index );
	
	else
		normal.x =  (( Height( index ) - Height( index + 1 ) ) +
					 ( Height( index - 1 ) - Height( index ) )) * 0.5;

	return normalize( normal );
}


//--------------------------------------------------------------------------------------
// Vertex Shaders
//--------------------------------------------------------------------------------------

VS_OUTPUT FilterHeightfieldVS( float2 vPos : POSITION )
{
	VS_OUTPUT Output = (VS_OUTPUT)0;  
	float4 position	 = 1.0f;
	position.xz		 = vPos * g_MetersPerVertex;

	// Pull the height from the buffer
	int index		 = ( vPos.y * g_NumVertsWide ) + vPos.x;
	position.y		 = g_Heights.Load( int2( index, 0 ) ) * g_MetersPerVertex;
	
	Output.Position  = mul(position, g_ViewProjectionMatrix);	

	// Compute the normal using a filter kernel
    float3 vNormalWorldSpace = FilterNormal(vPos, index);
    
    // Compute simple directional lighting equation
	float3 vTotalLightDiffuse	= g_LightDiffuse * max(0,dot(vNormalWorldSpace, g_LightDir));
	Output.Diffuse.rgb			= g_MaterialDiffuseColor * vTotalLightDiffuse;   
	Output.Diffuse.a			= 1.0f; 
       
    return Output; 
}

VS_OUTPUT RenderSceneVS( float3 vPos : POSITION, float3 vNormal : NORMAL, uint index : VERTEXID  )
{	
    // Transform the position from object space to homogeneous projection space
    VS_OUTPUT Output = (VS_OUTPUT)0;    
    float4 position  = 1.0f;
	position.xyz 	 = vPos;
    Output.Position  = mul(position, g_ViewProjectionMatrix);
    
    // Transform the normal from object space to world space    
    float3 vNormalWorldSpace = normalize(vNormal);
    
    // Compute simple directional lighting equation
	float3 vTotalLightDiffuse	= g_LightDiffuse * max(0,dot(vNormalWorldSpace, g_LightDir));
	Output.Diffuse.rgb			= g_MaterialDiffuseColor * vTotalLightDiffuse;   
	Output.Diffuse.a			= 1.0f; 
       
    return Output;     
}

VS_OUTPUT RenderNormalMapScene( float3 vPos : POSITION, float3 vNormal : NORMAL, uint index : VERTEXID )
{
    // Transform the position from object space to homogeneous projection space
    VS_OUTPUT Output = (VS_OUTPUT)0;    
    float4 position  = 1.0f;
	position.xyz 	 = vPos;
    Output.Position  = mul(position, g_ViewProjectionMatrix);	    

    // Transform the normal from object space to world space    
    float3 vNormalWorldSpace = normalize( ComputeNormal( index ) );
    
    // Compute simple directional lighting equation
	float3 vTotalLightDiffuse	= g_LightDiffuse * max(0,dot(vNormalWorldSpace, g_LightDir));
	Output.Diffuse.rgb			= g_MaterialDiffuseColor * vTotalLightDiffuse;   
	Output.Diffuse.a			= 1.0f; 
       
    return Output;     
}

GS_INPUT PassThroughVS( float3 vPos : POSITION, float3 vNormal : NORMAL, uint index : VERTEXID )
{    
    GS_INPUT Output = (GS_INPUT)0;     
	Output.Position.xyz = vPos;	
    return Output;     
}

//--------------------------------------------------------------------------------------
// Geometry Shaders
//--------------------------------------------------------------------------------------

[maxvertexcount(1)] 
void SurfaceNormalGS( triangle GS_INPUT input[3], inout PointStream<GS_INPUT> PStream )
{
	// Create a temp vector
	GS_INPUT Output = (GS_INPUT)0;

	// Compute the cross product and normalize to get the surface normal
	float3 edge1 = input[1].Position - input[0].Position;
	float3 edge2 = input[2].Position - input[0].Position;
	Output.Position.xyz = normalize( cross( edge2, edge1 ) );

	// Add the Point (with the normal) back to the PointStream to send to the buffer
	PStream.Append(Output);
}

//--------------------------------------------------------------------------------------
// Pixel Shaders
//--------------------------------------------------------------------------------------

PS_OUTPUT RenderScenePS( VS_OUTPUT In ) 
{ 
    PS_OUTPUT Output;
	Output.RGBColor = In.Diffuse;
    return Output;
}

//--------------------------------------------------------------------------------------
// D3D10 Techniques
//--------------------------------------------------------------------------------------

technique10 FilterSimpleTech
{
    pass P0
    {
		SetVertexShader( CompileShader( vs_4_0, FilterHeightfieldVS() ) );
		SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, RenderScenePS() ) );
		SetDepthStencilState( EnableDepth, 0 );
    }
}

technique10 MeshSimpleTech
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, RenderSceneVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, RenderScenePS() ) );
        SetDepthStencilState( EnableDepth, 0 );
    }
}

technique10 MeshWithNormalMapTech
{
    pass P0
    {
		SetVertexShader( CompileShader( vs_4_0, RenderNormalMapScene() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, RenderScenePS() ) );
        SetDepthStencilState( EnableDepth, 0 );
    }
}

GeometryShader gsNormalBuffer = ConstructGSWithSO( CompileShader( gs_4_0, SurfaceNormalGS() ), "POSITION.xyzw" );
technique10 MeshWithNormalMapSOTech
{
	pass P0
	{
		SetVertexShader( CompileShader( vs_4_0, PassThroughVS() ) );
		SetGeometryShader( gsNormalBuffer );
		SetPixelShader( NULL );
	}
	pass P1
	{
		SetVertexShader( CompileShader( vs_4_0, RenderNormalMapScene() ) );
		SetGeometryShader( NULL );
		SetPixelShader( CompileShader( ps_4_0, RenderScenePS() ) );
		SetDepthStencilState( EnableDepth, 0 );
	}
}