//-------------------------------------------------------------------------------------------------
//	Copyright  2008 Jeromy Walsh
//
//  You are hereby authorized to modify and/or redistribute the code contained within this file
//	so long as this header stays intact.  Any bugfixes or modifications you make to this code
//	can be sent to dragonwalsh (at) yahoo (dot) com for inclusion in the "official" release.
//-------------------------------------------------------------------------------------------------

#include "DXUT.h"
#include "heightfield.h"

//-------------------------------------------------------------------------------------------------
// [Methods]
//-------------------------------------------------------------------------------------------------

Heightfield::Heightfield( void ) 
{
	m_Width						= 16; 
	m_Depth						= 16;
	m_NumVertsDeep				= 0;
	m_NumVertsWide				= 0;
	m_MetersPerVertex			= 1.0f;	
	IsWireframe					= false;

	m_pHeights					= NULL;
	m_pVertices					= NULL;
	m_pIndices					= NULL;
	m_pSurfaceNormals			= NULL;

	m_pDevice					= NULL;
	m_pEffect					= NULL;
	m_pIndexBuffer				= NULL;
	m_pMeshVB					= NULL;
	m_pMeshIL					= NULL;
	m_pHeightfieldVB			= NULL;
	m_pHeightfieldIL			= NULL;
	m_pRasterizerState			= NULL;
	m_pWireframeState			= NULL;

	m_pFilterSimpleTech			= NULL;
	m_pMeshSimpleTech			= NULL;
	m_pMeshWithNormalMapTech	= NULL;
	m_pMeshWithNormalMapSOTech	= NULL;

	m_pMaterialDiffuseColor		= NULL;
	m_pMetersPerVertex			= NULL;
	m_pNumVertsWide				= NULL;
	m_pNumVertsDeep				= NULL;
	m_pSurfaceNormalsRV			= NULL;
	m_pHeightsRV				= NULL;
	
	m_pHeightBuffer				= NULL;
	m_pHeightBufferRV			= NULL;
	m_pNormalBuffer				= NULL;	
	m_pNormalBufferRV			= NULL;
	m_pNormalBufferSO			= NULL;
	m_pNormalBufferRVSO			= NULL;	
	
	Mode = NormalMode_CpuNormals;
}

Heightfield::~Heightfield() 
{
	Clear();
}

void Heightfield::Clear()
{
	SAFE_DELETE_ARRAY(m_pHeights);
	SAFE_DELETE_ARRAY(m_pIndices);
	SAFE_DELETE_ARRAY(m_pVertices);
	SAFE_DELETE_ARRAY(m_pSurfaceNormals);
	Release();
}

void Heightfield::Release()
{
	SAFE_RELEASE(m_pDevice);
	SAFE_RELEASE(m_pEffect);
	SAFE_RELEASE(m_pIndexBuffer);
	SAFE_RELEASE(m_pHeightfieldVB);
	SAFE_RELEASE(m_pHeightfieldIL);
	SAFE_RELEASE(m_pMeshVB);
	SAFE_RELEASE(m_pMeshIL);
	SAFE_RELEASE(m_pRasterizerState);
	SAFE_RELEASE(m_pWireframeState);
	SAFE_RELEASE(m_pHeightBuffer);
	SAFE_RELEASE(m_pHeightBufferRV);
	SAFE_RELEASE(m_pNormalBuffer);
	SAFE_RELEASE(m_pNormalBufferRV);
	SAFE_RELEASE(m_pNormalBufferSO);
	SAFE_RELEASE(m_pNormalBufferRVSO);
}

void Heightfield::Init( float resolution, ID3D10Device* pDevice, ID3D10Effect* pEffect  )
{
	Clear();

	// Claim the device
	m_pDevice = pDevice;
	m_pDevice->AddRef();

	// Claim the effect
	m_pEffect = pEffect;
	m_pEffect->AddRef();

	m_MetersPerVertex	= resolution;
	m_NumVertsWide		= (int)( m_Width / m_MetersPerVertex ) + 1;
	m_NumVertsDeep		= (int)( m_Depth / m_MetersPerVertex ) + 1;
	int numVertices		= m_NumVertsWide * m_NumVertsDeep; 
	int numRows			= m_NumVertsDeep - 1;
	int numColumns		= m_NumVertsWide - 1;
	int numSurfaces     = numColumns * numRows * 2;

	m_pHeights	= new float[ numVertices ];
	m_pVertices = new MeshVertex[ numVertices ];

	ZeroMemory(m_pHeights,  numVertices * sizeof(float) );
	ZeroMemory(m_pVertices, numVertices * sizeof(float) );

	m_pSurfaceNormals = new D3DXVECTOR3[numSurfaces];

	D3D10_RASTERIZER_DESC rasterDesc;
	ZeroMemory(&rasterDesc, sizeof(D3D10_RASTERIZER_DESC));
	rasterDesc.FillMode					= D3D10_FILL_SOLID;
	rasterDesc.CullMode					= D3D10_CULL_BACK;
	rasterDesc.FrontCounterClockwise	= true;
	rasterDesc.DepthClipEnable			= true;
	V(pDevice->CreateRasterizerState(&rasterDesc, &m_pRasterizerState));
	rasterDesc.FillMode					= D3D10_FILL_WIREFRAME;
	V(pDevice->CreateRasterizerState(&rasterDesc, &m_pWireframeState));

	m_pFilterSimpleTech			 = m_pEffect->GetTechniqueByName( "FilterSimpleTech" );
	m_pMeshSimpleTech			 = m_pEffect->GetTechniqueByName( "MeshSimpleTech" );
	m_pMeshWithNormalMapTech	 = m_pEffect->GetTechniqueByName( "MeshWithNormalMapTech" );
	m_pMeshWithNormalMapSOTech   = m_pEffect->GetTechniqueByName( "MeshWithNormalMapSOTech" );

    m_pMaterialDiffuseColor		 = m_pEffect->GetVariableByName( "g_MaterialDiffuseColor" )->AsVector();
	m_pMetersPerVertex			 = m_pEffect->GetVariableByName( "g_MetersPerVertex" )->AsScalar();
	m_pNumVertsDeep				 = m_pEffect->GetVariableByName( "g_NumVertsDeep" )->AsScalar();
	m_pNumVertsWide				 = m_pEffect->GetVariableByName( "g_NumVertsWide" )->AsScalar();
	m_pSurfaceNormalsRV			 = m_pEffect->GetVariableByName( "g_SurfaceNormals" )->AsShaderResource();
	m_pHeightsRV				 = m_pEffect->GetVariableByName( "g_Heights" )->AsShaderResource();

	D3DXCOLOR vBlue( 0, 0, 1, 1 );
	V( m_pMaterialDiffuseColor->SetFloatVector( ( float* )&vBlue ) );

	CreateIndexBuffer(m_NumVertsWide, numRows);
	CreateVertexBuffers(numVertices); 
	CreateShaderResources(numSurfaces);

	GenerateMeshData();
	ComputeNormals();            
	FillMeshVB();
}


void Heightfield::CreateShaderResources( int numSurfaces )
{
	// Create the non-streamed Shader Resources
	D3D10_BUFFER_DESC desc;
	D3D10_SHADER_RESOURCE_VIEW_DESC SRVDesc;

	// Create the height buffer for the filter method
	ZeroMemory( &desc, sizeof( D3D10_BUFFER_DESC ));
	ZeroMemory( &SRVDesc, sizeof( SRVDesc ) );
	desc.ByteWidth				= m_NumVertsDeep * m_NumVertsWide * sizeof(float);
	desc.Usage					= D3D10_USAGE_DYNAMIC;
	desc.BindFlags				= D3D10_BIND_SHADER_RESOURCE;
	desc.CPUAccessFlags 		= D3D10_CPU_ACCESS_WRITE;
	SRVDesc.Format				= DXGI_FORMAT_R32_FLOAT;
	SRVDesc.ViewDimension		= D3D10_SRV_DIMENSION_BUFFER;
	SRVDesc.Buffer.ElementWidth = m_NumVertsDeep * m_NumVertsWide;
	V( m_pDevice->CreateBuffer( &desc, NULL, &m_pHeightBuffer ) );
	V( m_pDevice->CreateShaderResourceView( m_pHeightBuffer, &SRVDesc, 
		&m_pHeightBufferRV ) );

	// Create the normal buffer for the GPU method
	ZeroMemory( &desc, sizeof( D3D10_BUFFER_DESC ));
	ZeroMemory( &SRVDesc, sizeof( SRVDesc ) );
	desc.ByteWidth				= numSurfaces * sizeof(D3DXVECTOR3);
	desc.Usage					= D3D10_USAGE_DYNAMIC;
	desc.BindFlags				= D3D10_BIND_SHADER_RESOURCE;
	desc.CPUAccessFlags 		= D3D10_CPU_ACCESS_WRITE;
	SRVDesc.Format				= DXGI_FORMAT_R32G32B32_FLOAT;
	SRVDesc.ViewDimension		= D3D10_SRV_DIMENSION_BUFFER;
	SRVDesc.Buffer.ElementWidth = numSurfaces;
	V( m_pDevice->CreateBuffer( &desc, NULL, &m_pNormalBuffer ) );
	V( m_pDevice->CreateShaderResourceView( m_pNormalBuffer, &SRVDesc, 
		&m_pNormalBufferRV ) );

	// Create output normal buffer for the Stream Output
	ZeroMemory( &desc, sizeof( D3D10_BUFFER_DESC ));
	ZeroMemory( &SRVDesc, sizeof( SRVDesc ) );
	desc.ByteWidth				= numSurfaces * sizeof( D3DXVECTOR4 );
	desc.Usage					= D3D10_USAGE_DEFAULT;
	desc.BindFlags				= D3D10_BIND_SHADER_RESOURCE | D3D10_BIND_STREAM_OUTPUT;
	SRVDesc.Format				= DXGI_FORMAT_R32G32B32A32_FLOAT;
	SRVDesc.ViewDimension		= D3D10_SRV_DIMENSION_BUFFER;
	SRVDesc.Buffer.ElementWidth = numSurfaces;
	V( m_pDevice->CreateBuffer( &desc, NULL, &m_pNormalBufferSO ) );
	V( m_pDevice->CreateShaderResourceView( m_pNormalBufferSO, &SRVDesc, 
		&m_pNormalBufferRVSO ) );
}


void Heightfield::Update( float elapsedTime )
{
	// Compute the starting offset within the sin() function
	static float offset = 0;
	offset += 5.0f * elapsedTime;

	// Compute the delta between vertices
	float delta	 = 16.0 / m_NumVertsWide;
	float height = 0.0f;
	int index;
	int maxValue;
	
	if( offset >= 360.0f ) offset -= 360.0f;

	for( int z = 0; z < m_NumVertsDeep; z++)
	{
		for( int x = 0; x < m_NumVertsWide; x++)
		{
			// Determine where in the sine curve we are
			maxValue = x > z ? x : z;
			height = sin( maxValue * delta + offset );

			// Divide by mpv, so when we scale later it doesn't shrink
			// our curve
			m_pHeights[z*m_NumVertsWide+x] = height / m_MetersPerVertex;
		}
	}

	if( Mode == NormalMode_FilterNormals )
		FillFilterBuffer();
	else
	{
		GenerateMeshData();
		ComputeNormals();            
		FillMeshVB(); 
	}
}


void Heightfield::GenerateMeshData()
{
	unsigned i = 0;
    for (int z = 0; z < m_NumVertsDeep; z++)
    {
        for (int x = 0; x < m_NumVertsWide; x++)
        {
			m_pVertices[i].pos.x = (float)x * m_MetersPerVertex;
			m_pVertices[i].pos.z = (float)z * m_MetersPerVertex;
			m_pVertices[i].pos.y = m_pHeights[i] * m_MetersPerVertex;
			m_pVertices[i].i = i;
			i++;
        }
    }
}


void Heightfield::FillFilterBuffer()
{
	// Map the buffer
	int bufferSize = (m_NumVertsWide * m_NumVertsDeep) * sizeof(float);

	float* pHeights = NULL; 
	V(m_pHeightBuffer->Map(D3D10_MAP_WRITE_DISCARD, 0, (void**)&pHeights));

	// Copy the height data to the buffer
	memcpy(pHeights, m_pHeights, bufferSize );

	// Unmap the buffer
	m_pHeightBuffer->Unmap();
}


void Heightfield::Draw()
{
	// Init some locals
	ID3D10EffectTechnique* pTechnique = NULL;
	int numRows     = m_NumVertsDeep - 1;
	int numIndices	= 2 * m_NumVertsWide;
	UINT offset		= 0;
	UINT stride		= 0;

	V( m_pNumVertsDeep->SetInt( m_NumVertsDeep ) );
	V( m_pNumVertsWide->SetInt( m_NumVertsWide ) );
	V( m_pMetersPerVertex->SetFloat( m_MetersPerVertex ) );

	// Set the Raster states necessary for rendering this Heightfield
	ID3D10RasterizerState *pState;
	m_pDevice->RSGetState(&pState);

	if( !IsWireframe )
		m_pDevice->RSSetState( m_pRasterizerState );
	else
		m_pDevice->RSSetState( m_pWireframeState );

	// Set up the IA state
	m_pDevice->IASetPrimitiveTopology(D3D10_PRIMITIVE_TOPOLOGY_TRIANGLESTRIP);
	m_pDevice->IASetIndexBuffer(m_pIndexBuffer, DXGI_FORMAT_R32_UINT, 0 );

	// Finish initializing the IA stage for the heightfield 
	// and set the technique depending upon the mode
	if( Mode == NormalMode_FilterNormals )
	{
		stride	= sizeof( FilterVertex );
		m_pDevice->IASetInputLayout(m_pHeightfieldIL);
		m_pDevice->IASetVertexBuffers(0, 1, &m_pHeightfieldVB, &stride, &offset);
		V(m_pHeightsRV->SetResource( m_pHeightBufferRV ));
		pTechnique = m_pFilterSimpleTech;
	}
	else
	{
		stride	= sizeof( MeshVertex );
		ID3D10ShaderResourceView* pViews[] = { NULL };
		m_pDevice->IASetInputLayout(m_pMeshIL);
		m_pDevice->IASetVertexBuffers(0, 1, &m_pMeshVB, &stride, &offset);
		
		switch(Mode)
		{
			case NormalMode_CpuNormals:
				pTechnique = m_pMeshSimpleTech;
				break;

			case NormalMode_GpuNormals:
				V(m_pSurfaceNormalsRV->SetResource( m_pNormalBufferRV ));
				pTechnique = m_pMeshWithNormalMapTech;
				break;

			case NormalMode_StreamNormals:
				m_pDevice->VSSetShaderResources(0, 1, pViews );
				m_pDevice->SOSetTargets(1, &m_pNormalBufferSO, &offset );
				V(m_pSurfaceNormalsRV->SetResource( m_pNormalBufferRVSO ));
				pTechnique = m_pMeshWithNormalMapSOTech;
				break;
		}
	}

	D3D10_TECHNIQUE_DESC desc;
	pTechnique->GetDesc(&desc);

	for( unsigned i = 0; i < desc.Passes; i++ )
	{
		pTechnique->GetPassByIndex(i)->Apply(0);
		for (int j = 0; j < numRows; j++)
			m_pDevice->DrawIndexed(numIndices, j * numIndices, 0 );

		// Clear the Output Stream before pass 1 to unbind the buffer
		m_pDevice->SOSetTargets(0, NULL, &offset );
	}	
	
	m_pDevice->RSSetState( pState );
	pState->Release();
}

void Heightfield::CreateVertexBuffers( int numVertices )
{
	// Allocate vertices for static vertex buffer
	D3DXVECTOR2* pVertices = new D3DXVECTOR2[numVertices];
	InitHeightfieldData(pVertices);

	// Create the heightfield's vertex input layout
	const D3D10_INPUT_ELEMENT_DESC heightfieldLayout[] =
	{
		{ "POSITION",  0, DXGI_FORMAT_R32G32_FLOAT, 0, 0,  D3D10_INPUT_PER_VERTEX_DATA, 0 },
	};

	// Create the mesh's input layout
    const D3D10_INPUT_ELEMENT_DESC meshLayout[] =
    {
        { "POSITION",  0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0,  D3D10_INPUT_PER_VERTEX_DATA, 0 },
        { "NORMAL",    0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 12, D3D10_INPUT_PER_VERTEX_DATA, 0 },
		{ "VERTEXID",  0, DXGI_FORMAT_R32_UINT,        0, 24, D3D10_INPUT_PER_VERTEX_DATA, 0 },
    };
	
	// Create the Heightfield's Input Layout
	D3D10_PASS_DESC PassDesc;
	V( m_pFilterSimpleTech->GetPassByIndex(0)->GetDesc(&PassDesc) );
	V( m_pDevice->CreateInputLayout( heightfieldLayout, 1, PassDesc.pIAInputSignature,
        PassDesc.IAInputSignatureSize, &m_pHeightfieldIL ) );

	// Create the Mesh IL
    V( m_pMeshSimpleTech->GetPassByIndex( 0 )->GetDesc( &PassDesc ) );
    V( m_pDevice->CreateInputLayout( meshLayout, 3, PassDesc.pIAInputSignature,
        PassDesc.IAInputSignatureSize, &m_pMeshIL ) );

	D3D10_BUFFER_DESC desc;
	desc = CD3D10_BUFFER_DESC(numVertices * sizeof( FilterVertex ), 
		D3D10_BIND_VERTEX_BUFFER, D3D10_USAGE_DEFAULT );
	D3D10_SUBRESOURCE_DATA subData = { pVertices, 0, 0 };
	V(m_pDevice->CreateBuffer(&desc, &subData, &m_pHeightfieldVB ));

	desc = CD3D10_BUFFER_DESC(numVertices * sizeof( MeshVertex ), 
		D3D10_BIND_VERTEX_BUFFER, D3D10_USAGE_DYNAMIC, D3D10_CPU_ACCESS_WRITE );
	V(m_pDevice->CreateBuffer(&desc, NULL, &m_pMeshVB ));

	delete [] pVertices; // Clean up the vertices after we're done
}

void Heightfield::CreateIndexBuffer(int numVertsWide, int numRows)
{
	int numIndices = 2 * numVertsWide * numRows;
	m_pIndices = new int[ numIndices ];

	int i = 0;
	for (int row = 0; row < numRows; row++)
    {
        for (int column = 0; column < numVertsWide; column++)
        {
            m_pIndices[i++] = (row + 1) * numVertsWide + column; // Bottom Vertex
            m_pIndices[i++] = row * numVertsWide + column; // Top vertex
        }
    }

	// Store the data in subresource struct, then create the buffer
	// description to be used when creating the buffer
	D3D10_SUBRESOURCE_DATA data = { m_pIndices, 0, 0 };
	D3D10_BUFFER_DESC desc =
	{
		numIndices * sizeof(int),
		D3D10_USAGE_DEFAULT,
		D3D10_BIND_INDEX_BUFFER,
		0, 0
	};

	V(m_pDevice->CreateBuffer(&desc, &data, &m_pIndexBuffer ));
}

void Heightfield::InitHeightfieldData(D3DXVECTOR2* pData)
{
	int i = 0;
    for (int z = 0; z < m_NumVertsDeep; z++)
    {
        for (int x = 0; x < m_NumVertsWide; x++)
        {
			pData[i].x	 = (float)x;
			pData[i++].y = (float)z;
        }
    }
}

void Heightfield::FillMeshVB()
{
	MeshVertex* pVertices;
	int numVertices = m_NumVertsDeep * m_NumVertsWide;
    
	V(m_pMeshVB->Map(D3D10_MAP_WRITE_DISCARD, 0, (void**)&pVertices ));
	memcpy(pVertices, m_pVertices, numVertices * sizeof(MeshVertex) );
	m_pMeshVB->Unmap();
}

void Heightfield::ComputeNormals()
{
	if( Mode != NormalMode_CpuNormals && Mode != NormalMode_GpuNormals)
		return;

	ComputeSurfaceNormals();

	if( Mode == NormalMode_CpuNormals )
		ComputeCPUNormals();
	else
		ComputeGPUNormals();
}

void Heightfield::ComputeSurfaceNormals()
{
	int i = 0;
	int numRows			= m_NumVertsDeep - 1;
	int indicesPerRow   = m_NumVertsWide * 2;
	D3DXVECTOR3 line1, line2;
	
	for( int j = 0; j < numRows; j++ )
	{
		int startIndex	= (j * indicesPerRow);
		int endIndex	= (j + 1) * indicesPerRow - 3;

		for (int index = startIndex; index < endIndex; index++)
        {
			// Top-Left Triangle
			line1 = m_pVertices[ m_pIndices[index] ].pos   - 
					m_pVertices[ m_pIndices[index+1]].pos;
			line2 = m_pVertices[ m_pIndices[index+2] ].pos - 
					m_pVertices[ m_pIndices[index+1] ].pos;
			D3DXVec3Cross( &m_pSurfaceNormals[i++], &line1, &line2 );
			index++;

			// Bottom-Right Triangle
			line1 = m_pVertices[ m_pIndices[index] ].pos   - 
					m_pVertices[ m_pIndices[index+1]].pos;
			line2 = m_pVertices[ m_pIndices[index+2] ].pos - 
					m_pVertices[ m_pIndices[index+1] ].pos;
			D3DXVec3Cross( &m_pSurfaceNormals[i++], &line2, &line1 );
		}
	}
}

void Heightfield::ComputeGPUNormals()
{
	int bufferSize = 2 * (m_NumVertsWide - 1) * 
						 (m_NumVertsDeep - 1) * 
						  sizeof(D3DXVECTOR3);

	D3DXVECTOR3* pSurfaceNormals = NULL; 
	V(m_pNormalBuffer->Map(D3D10_MAP_WRITE_DISCARD, 0, (void**)&pSurfaceNormals));
	memcpy(pSurfaceNormals, m_pSurfaceNormals, bufferSize );
	m_pNormalBuffer->Unmap();
}


void Heightfield::ComputeCPUNormals()
{
	// Compute the vertex normals
	int numRows			= m_NumVertsDeep - 1;
	int numColumns		= m_NumVertsWide - 1;    
	int maxIndex		= m_NumVertsWide * numRows;
	int surfacesPerRow  = numColumns * 2;	
	int numSurfaces     = surfacesPerRow * numRows;
	int topRight		= m_NumVertsWide * m_NumVertsDeep - 1;
	int offset;

	// handle the bottom left
	m_pVertices[0].norm = m_pSurfaceNormals[0] + m_pSurfaceNormals[1];
	int vertIndex		= 1;

	// Bottom Row
	for (; vertIndex < numColumns; vertIndex++)
	{
		// handle the top center
		int baseIndex = vertIndex * 2;
		m_pVertices[vertIndex].norm = 
			m_pSurfaceNormals[baseIndex - 1] +
			m_pSurfaceNormals[baseIndex] +
			m_pSurfaceNormals[baseIndex + 1];
	}

	 // handle the bottom right
	 m_pVertices[vertIndex].norm = m_pSurfaceNormals[numColumns * 2 - 1];
	 vertIndex++;

	// Middle
	for (; vertIndex < maxIndex; vertIndex++)
	{
		offset = vertIndex % m_NumVertsWide;

		// Is the vertex along the left edge
		if (offset == 0)
		{
			int baseIndex = (vertIndex - m_NumVertsWide) / m_NumVertsWide * surfacesPerRow;
			m_pVertices[vertIndex].norm = 
				m_pSurfaceNormals[baseIndex] +
				m_pSurfaceNormals[baseIndex + surfacesPerRow] +
				m_pSurfaceNormals[baseIndex + surfacesPerRow + 1];
		}

		// Is the vertex along the right edge
		else if (offset == numColumns)
		{
			int baseIndex = ((vertIndex + 1) / m_NumVertsWide * surfacesPerRow) - 1;
			m_pVertices[vertIndex].norm = 
				m_pSurfaceNormals[baseIndex] +
				m_pSurfaceNormals[baseIndex - surfacesPerRow] +
				m_pSurfaceNormals[baseIndex - (surfacesPerRow + 1)];
		}
		else
		{
			int baseIndex = ((vertIndex - m_NumVertsWide) / m_NumVertsWide * surfacesPerRow) +
				2 * ((vertIndex % m_NumVertsWide) - 1);
			int offset = baseIndex + surfacesPerRow + 1;
			m_pVertices[vertIndex].norm =
				m_pSurfaceNormals[baseIndex] +
				m_pSurfaceNormals[baseIndex + 1] +
				m_pSurfaceNormals[baseIndex + 2] +
				m_pSurfaceNormals[offset] +
				m_pSurfaceNormals[offset + 1] +
				m_pSurfaceNormals[offset + 2];
		}
	}

	// handle the top left
	m_pVertices[vertIndex].norm = m_pSurfaceNormals[(numColumns * 2) * (numRows - 1)];
	vertIndex++;

	// top row
	for (; vertIndex < topRight; vertIndex++)
	{
		// handle the bottm center
		int baseIndex = ((vertIndex - m_NumVertsWide) / m_NumVertsWide * surfacesPerRow) +
						 2 * ((vertIndex % m_NumVertsWide) - 1);
		m_pVertices[vertIndex].norm =
			m_pSurfaceNormals[baseIndex] +
			m_pSurfaceNormals[baseIndex + 1] +
			m_pSurfaceNormals[baseIndex + 2];
	}

	// handle the top right
	m_pVertices[vertIndex].norm = 
		m_pSurfaceNormals[numSurfaces - 1] +
		m_pSurfaceNormals[numSurfaces - 2];
}